#include "eggservice.h"

#include <avahi-client/client.h>
#include <avahi-client/publish.h>
#include <avahi-glib/glib-watch.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <string.h>
#include <stdlib.h>

enum
{
  PROP_NONE,
  PROP_DESCRIPTION,
  PROP_SERVICE_TYPE,
  PROP_ADDRESS
};

struct _EggServicePrivate
{
  gchar *address;
  gchar *description;
  gchar *service_type;

  AvahiClient *avahi;
  AvahiEntryGroup *entries;
  AvahiGLibPoll *poll;

  GHashTable *objects;
  DBusServer *server;
};

G_DEFINE_TYPE (EggService, egg_service, G_TYPE_OBJECT)

static void
egg_service_announce (gpointer key,
                      gpointer value,
                      gpointer data)
{
  g_print ("registering %s object at %s\n", 
           G_OBJECT_TYPE_NAME (value),
           (const gchar*) key);

  dbus_g_connection_register_g_object (data, key, value);
}

static void
egg_service_new_connection_cb (DBusServer     *server G_GNUC_UNUSED,
                               DBusConnection *connection,
                               void           *data)
{
  EggService *self = data;

  dbus_connection_ref (connection);
  dbus_connection_setup_with_g_main (connection, NULL);
  dbus_connection_set_allow_anonymous (connection, TRUE);

  g_hash_table_foreach (self->priv->objects, egg_service_announce,
                        dbus_connection_get_g_connection (connection));
}

static void
egg_service_entries_cb (AvahiEntryGroup      *group G_GNUC_UNUSED,
                        AvahiEntryGroupState  state G_GNUC_UNUSED, 
                        void                 *data)
{
  EggService *self G_GNUC_UNUSED = data;
}

static void
egg_service_startup_server (EggService *self)
{
  const char *auth_mechanisms[] = { "ANONYMOUS", NULL };
  char *address = self->priv->address;
  int i, n_entries;

  DBusAddressEntry **entries;
  DBusError error;

  g_return_if_fail (NULL == self->priv->server);

  if (NULL == address)
    {
      const gchar *host = avahi_client_get_host_name_fqdn (self->priv->avahi);
      address = g_strdup_printf ("tcp:host=%s,all_interfaces=true", host);
    }

  dbus_error_init (&error);

  self->priv->server = dbus_server_listen (address, &error);

  if (!self->priv->server)
    {
      g_warning ("Cannot start DBus server: %s", error.message);
      goto out;
    }

  dbus_server_setup_with_g_main (self->priv->server, NULL);
  dbus_server_set_auth_mechanisms (self->priv->server, auth_mechanisms);
  dbus_server_set_new_connection_function (self->priv->server,
                                           egg_service_new_connection_cb,
                                           self, NULL);

  if (address != self->priv->address)
    g_free (address);

  address = dbus_server_get_address (self->priv->server);
  g_file_set_contents ("server-address.txt", address, -1, NULL);
  
  if (dbus_parse_address (address, &entries, &n_entries, NULL))
    { 
      const char *description = self->priv->description;
      const char *service_type = self->priv->service_type;

      if (NULL == description)
        description = g_get_application_name ();
      if (NULL == service_type)
        service_type = "_dbus._tcp";

g_debug ("%s/%s", description, service_type);

      for (i = 0; i < n_entries; ++i)
        {
          const gchar *host, *port;

          if (strcmp ("tcp", dbus_address_entry_get_method (entries[i])) ||
              NULL == (host = dbus_address_entry_get_value (entries[i], "host")) ||
              NULL == (port = dbus_address_entry_get_value (entries[i], "port")))
            continue;

          if (NULL == self->priv->entries)
            self->priv->entries = avahi_entry_group_new (self->priv->avahi,
                                                         egg_service_entries_cb,
                                                         self);

          avahi_entry_group_add_service (self->priv->entries, 
                                         AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC, 
                                         0, description, service_type, NULL,
                                         host, atoi (port), NULL);
        }

      if (self->priv->entries)
        avahi_entry_group_commit (self->priv->entries);

      dbus_address_entries_free (entries);
    }

out:
  if (address != self->priv->address)
    g_free (address);

  dbus_error_free (&error);
}

static void
egg_service_avahi_cb (AvahiClient      *avahi,
                      AvahiClientState  state, 
                      void             *data)
{
  EggService *self = data;
  self->priv->avahi = avahi;

  switch (state)
    {
      case AVAHI_SERVER_RUNNING:
        egg_service_startup_server (self);
        break;

      case AVAHI_SERVER_COLLISION:
      case AVAHI_SERVER_REGISTERING:
        g_debug ("stop service...");
        break;

      default:
        break;
    }
}

static void
egg_service_init (EggService *self)
{
  self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self, EGG_TYPE_SERVICE, EggServicePrivate);

  g_assert (NULL != g_get_application_name ());

  self->priv->poll = avahi_glib_poll_new (NULL, G_PRIORITY_DEFAULT);
  self->priv->objects = g_hash_table_new_full (g_str_hash, g_str_equal,
                                               g_free, g_object_unref);
}

static void
egg_service_get_property (GObject    *object,
                          guint       prop_id,
                          GValue     *value,
                          GParamSpec *pspec)
{
  EggService *self = EGG_SERVICE (object);

  switch (prop_id)
    {
      case PROP_DESCRIPTION:
        g_value_set_string (value, self->priv->description);
        break;

      case PROP_SERVICE_TYPE:
        g_value_set_string (value, self->priv->service_type);
        break;

      case PROP_ADDRESS:
        if (self->priv->server)
          {
            gchar *address = dbus_server_get_address (self->priv->server);
            g_value_take_string (value, address);
          }
        else
          g_value_set_string (value, self->priv->address);

        break;

      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
egg_service_set_property (GObject      *object,
                          guint         prop_id,
                          const GValue *value,
                          GParamSpec   *pspec)
{
  EggService *self = EGG_SERVICE (object);

  switch (prop_id)
    {
      case PROP_DESCRIPTION:
        self->priv->description = g_value_dup_string (value);
        break;

      case PROP_SERVICE_TYPE:
        self->priv->service_type = g_value_dup_string (value);
        break;

      case PROP_ADDRESS:
        g_assert (NULL == self->priv->address);
        self->priv->address = g_value_dup_string (value);
        break;

      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
egg_service_dispose (GObject *object)
{
  EggService *self = EGG_SERVICE (object);

  egg_service_stop (self);

  if (self->priv->poll)
    {
      avahi_glib_poll_free (self->priv->poll);
      self->priv->poll = NULL;
    }

  if (self->priv->objects)
    {
      g_hash_table_unref (self->priv->objects);
      self->priv->objects = NULL;
    }

  g_free (self->priv->address);
  self->priv->address = NULL;

  g_free (self->priv->description);
  self->priv->description = NULL;

  g_free (self->priv->service_type);
  self->priv->service_type = NULL;

  G_OBJECT_CLASS (egg_service_parent_class)->dispose (object);
}

static void
egg_service_class_init (EggServiceClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->get_property = egg_service_get_property;
  object_class->set_property = egg_service_set_property;
  object_class->dispose = egg_service_dispose;

  g_object_class_install_property (object_class, PROP_DESCRIPTION,
                                   g_param_spec_string ("description", "Description", 
                                                        "Description of this service", NULL,
                                                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT |
                                                        G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
                                                        G_PARAM_STATIC_BLURB));
  g_object_class_install_property (object_class, PROP_SERVICE_TYPE,
                                   g_param_spec_string ("service-type", "Service Type", 
                                                        "Zeroconf service type", NULL,
                                                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT |
                                                        G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
                                                        G_PARAM_STATIC_BLURB));
  g_object_class_install_property (object_class, PROP_ADDRESS,
                                   g_param_spec_string ("address", "Address", 
                                                        "DBus address of this service", NULL,
                                                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT |
                                                        G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
                                                        G_PARAM_STATIC_BLURB));

  g_type_class_add_private (klass, sizeof (EggServicePrivate));
}

EggService*
egg_service_new (const char *description,
                 const char *service_type,
                 const char *address)
{
  return g_object_new (EGG_TYPE_SERVICE,
                       "description", description,
                       "service-type", service_type,
                       "address", address,
                       NULL);
}

static gboolean
egg_service_is_running (EggService *self)
{
  return (NULL != self->priv->avahi);
}

void
egg_service_start (EggService *self)
{
  g_return_if_fail (EGG_IS_SERVICE (self));

  if (egg_service_is_running (self))
    {
      g_warning ("The service is running already");
      return;
    }

  avahi_client_new (avahi_glib_poll_get (self->priv->poll),
                    AVAHI_CLIENT_NO_FAIL, egg_service_avahi_cb,
                    self, NULL);
}

void
egg_service_stop (EggService *self)
{
  g_return_if_fail (EGG_IS_SERVICE (self));

  if (egg_service_is_running (self))
    {
      if (self->priv->server)
        {
          dbus_server_disconnect (self->priv->server);
          dbus_server_unref (self->priv->server);
          self->priv->server = NULL;
        }

      if (self->priv->avahi)
        {
          avahi_client_free (self->priv->avahi);
          self->priv->avahi = NULL;
        }
    }
}

void
egg_service_publish (EggService *self,
                     const char *path,
                     GObject    *object)
{
  g_return_if_fail (EGG_IS_SERVICE (self));
  g_return_if_fail (G_IS_OBJECT (object));
  g_return_if_fail (NULL != path);

  if (egg_service_is_running (self))
    {
      g_warning ("Cannot publish services after starting the service");
      return;
    }

  g_hash_table_insert (self->priv->objects, g_strdup (path), g_object_ref (object));
}
